import * as THREE from '../build/three.module.js';
import Stats from './jsm/libs/stats.module.js';

class App {
    constructor() {
        this._initialize();
    }

    _initialize() {
        this.domWebGL = document.createElement('div');
        document.body.appendChild(this.domWebGL);

        let renderer = new THREE.WebGLRenderer({ antialias: true });
        this.domWebGL.appendChild(renderer.domElement);
        this._renderer = renderer;

        const scene = new THREE.Scene();
        scene.background = new THREE.Color(0x050505);
        this._scene = scene;

        this._setupCamera();
        this._setupModel();
        this._setupGUI();
        
        window.onresize = this.resize.bind(this);

        this.resize();

        this.render();
    }

    _setupCamera() {
        const camera = new THREE.PerspectiveCamera(40, window.innerWidth / window.innerHeight, 1, 10000);
        camera.position.z = 300;
        this._camera = camera;
    }

    _createShaderMaterial() {
        const uniforms = {
            "color": { value: new THREE.Color(0xffffff) },
            pointTexture: { value: new THREE.TextureLoader().load("textures/sprites/spark1.png")}
        };

        const vertexCode = `
            attribute float size;
            attribute vec3 customColor; 
            
            varying vec3 vColor;

            void main() {
                vColor = customColor;
                vec4 mvPosition = modelViewMatrix * vec4(position, 1.0);
                gl_PointSize = size * (300.0 / -mvPosition.z);
                gl_Position = projectionMatrix * mvPosition;
            }
        `;

        const fragmentCode = `
            uniform vec3 color;
            uniform sampler2D pointTexture;

            varying vec3 vColor;

            void main() {
                gl_FragColor = vec4(color * vColor, 1.0);
                gl_FragColor = gl_FragColor * texture2D(pointTexture, gl_PointCoord);
            }
        `;
        
        const shaderMaterial = new THREE.ShaderMaterial({
            uniforms: uniforms,
            vertexShader: vertexCode,
            fragmentShader: fragmentCode,
            blending: THREE.AdditiveBlending,
            depthTest: false,
            transparent: true
        });

        return shaderMaterial;
    }

    _setupModel(font) {
        const amount = 100000;
        const radius = 200;

        const positions = new Float32Array(amount * 3);
        const colors = new Float32Array(amount * 3);
        const sizes = new Float32Array(amount);

        const vertex = new THREE.Vector3();
        const color = new THREE.Color(0xffffff);

        for(let i=0; i<amount; i++) {
            vertex.x = (Math.random() * 2 - 1) * radius;
            vertex.y = (Math.random() * 2 - 1) * radius;
            vertex.z = (Math.random() * 2 - 1) * radius;
            vertex.toArray(positions, i*3);

            if(vertex.x < 0) {
                color.setHSL(0.5 + 0.1 * (i / amount), 0.7, 0.5);
            } else {
                color.setHSL(0.0 + 0.1 * (i / amount), 0.9, 0.5);
            }

            color.toArray(colors, i*3);
            sizes[i] = 10;
        }

        const geometry = new THREE.BufferGeometry();
        geometry.setAttribute("position", new THREE.BufferAttribute(positions, 3));
        geometry.setAttribute("customColor", new THREE.BufferAttribute(colors, 3));
        geometry.setAttribute("size", new THREE.BufferAttribute(sizes, 1));

        const material = this._createShaderMaterial();

        const sphere = new THREE.Points(geometry, material);
        this._scene.add(sphere);

        this._sphere = sphere;
    }

    update() {
        const time = Date.now() * 0.005;

        this._sphere.rotation.z = 0.01 * time;

        const geometry = this._sphere.geometry;
        const attributes = geometry.attributes;

        for(let i=0; i<attributes.size.array.length; i++) {
            attributes.size.array[i] = 14 + 13 * Math.sin(0.1 * i + time);
        }

        attributes.size.needsUpdate = true;        
    }

    render() {
        requestAnimationFrame(this.render.bind(this));
        
        this.update();
        this._renderer.render(this._scene, this._camera);
        this._stats.update();
    }

    resize() {
        this._camera.aspect = window.innerWidth / window.innerHeight;
        this._camera.updateProjectionMatrix();

        this._renderer.setPixelRatio(window.devicePixelRatio);
        this._renderer.setSize(window.innerWidth, window.innerHeight);
    }

    _setupGUI() {
        const stats = new Stats();
        this.domWebGL.appendChild(stats.dom);
        this._stats = stats;
    }
}

window.onload = function () {
    (new App()).render();
}